"use strict";
/**
 * Copyright (c) Microsoft Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RecorderSupplement = void 0;
const fs = __importStar(require("fs"));
const codeGenerator_1 = require("./recorder/codeGenerator");
const utils_1 = require("./recorder/utils");
const page_1 = require("../page");
const frames_1 = require("../frames");
const browserContext_1 = require("../browserContext");
const javascript_1 = require("./recorder/javascript");
const csharp_1 = require("./recorder/csharp");
const python_1 = require("./recorder/python");
const recorderSource = __importStar(require("../../generated/recorderSource"));
const consoleApiSource = __importStar(require("../../generated/consoleApiSource"));
const outputs_1 = require("./recorder/outputs");
const recorderApp_1 = require("./recorder/recorderApp");
const instrumentation_1 = require("../instrumentation");
const symbol = Symbol('RecorderSupplement');
class RecorderSupplement {
    constructor(context, params) {
        this._pageAliases = new Map();
        this._lastPopupOrdinal = 0;
        this._lastDialogOrdinal = 0;
        this._timers = new Set();
        this._resumeCallback = null;
        this._pauseDetails = null;
        this._recorderApp = null;
        this._callMetadata = null;
        this._pauseOnNextStatement = true;
        this._sourceCache = new Map();
        this._sdkObject = null;
        this._context = context;
        this._params = params;
        this._mode = params.startRecording ? 'recording' : 'none';
        let languageGenerator;
        const language = params.language || context._options.sdkLanguage;
        switch (language) {
            case 'javascript':
                languageGenerator = new javascript_1.JavaScriptLanguageGenerator();
                break;
            case 'csharp':
                languageGenerator = new csharp_1.CSharpLanguageGenerator();
                break;
            case 'python':
            case 'python-async':
                languageGenerator = new python_1.PythonLanguageGenerator(params.language === 'python-async');
                break;
            default: throw new Error(`Invalid target: '${params.language}'`);
        }
        let highlighterType = language;
        if (highlighterType === 'python-async')
            highlighterType = 'python';
        this._highlighterType = highlighterType;
        this._bufferedOutput = new outputs_1.BufferedOutput(async (text) => {
            if (this._recorderApp)
                this._recorderApp.setSource(text, highlighterType);
        });
        const outputs = [this._bufferedOutput];
        if (params.outputFile)
            outputs.push(new outputs_1.FileOutput(params.outputFile));
        this._output = new outputs_1.OutputMultiplexer(outputs);
        this._output.setEnabled(!!params.startRecording);
        context.on(browserContext_1.BrowserContext.Events.BeforeClose, () => this._output.flush());
        const generator = new codeGenerator_1.CodeGenerator(context._browser.options.name, !!params.startRecording, params.launchOptions || {}, params.contextOptions || {}, this._output, languageGenerator, params.device, params.saveStorage);
        this._generator = generator;
    }
    static getOrCreate(context, params = {}) {
        let recorderPromise = context[symbol];
        if (!recorderPromise) {
            const recorder = new RecorderSupplement(context, params);
            recorderPromise = recorder.install().then(() => recorder);
            context[symbol] = recorderPromise;
        }
        return recorderPromise;
    }
    async install() {
        const recorderApp = await recorderApp_1.RecorderApp.open(this._context);
        this._recorderApp = recorderApp;
        recorderApp.once('close', () => {
            this._recorderApp = null;
        });
        recorderApp.on('event', (data) => {
            if (data.event === 'setMode') {
                this._mode = data.params.mode;
                recorderApp.setMode(this._mode);
                this._output.setEnabled(this._mode === 'recording');
                if (this._mode !== 'none')
                    this._context.pages()[0].bringToFront().catch(() => { });
                return;
            }
            if (data.event === 'step') {
                this._resume(true);
                return;
            }
            if (data.event === 'resume') {
                this._resume(false);
                return;
            }
            if (data.event === 'pause') {
                this._pauseOnNextStatement = true;
                return;
            }
            if (data.event === 'clear') {
                this._clearScript();
                return;
            }
        });
        await Promise.all([
            recorderApp.setMode(this._mode),
            recorderApp.setPaused(this._pauseDetails),
            recorderApp.setSource(this._bufferedOutput.buffer(), this._highlighterType)
        ]);
        this._context.on(browserContext_1.BrowserContext.Events.Page, page => this._onPage(page));
        for (const page of this._context.pages())
            this._onPage(page);
        this._context.once(browserContext_1.BrowserContext.Events.Close, () => {
            for (const timer of this._timers)
                clearTimeout(timer);
            this._timers.clear();
            recorderApp.close().catch(() => { });
        });
        // Input actions that potentially lead to navigation are intercepted on the page and are
        // performed by the Playwright.
        await this._context.exposeBinding('_playwrightRecorderPerformAction', false, (source, action) => this._performAction(source.frame, action));
        // Other non-essential actions are simply being recorded.
        await this._context.exposeBinding('_playwrightRecorderRecordAction', false, (source, action) => this._recordAction(source.frame, action));
        // Commits last action so that no further signals are added to it.
        await this._context.exposeBinding('_playwrightRecorderCommitAction', false, (source, action) => this._generator.commitLastAction());
        await this._context.exposeBinding('_playwrightRecorderState', false, source => {
            var _a, _b, _c, _d;
            let actionPoint = undefined;
            let actionSelector = undefined;
            if (source.page === ((_b = (_a = this._sdkObject) === null || _a === void 0 ? void 0 : _a.attribution) === null || _b === void 0 ? void 0 : _b.page)) {
                actionPoint = (_c = this._callMetadata) === null || _c === void 0 ? void 0 : _c.point;
                actionSelector = (_d = this._callMetadata) === null || _d === void 0 ? void 0 : _d.params.selector;
            }
            const uiState = { mode: this._mode, actionPoint, actionSelector };
            return uiState;
        });
        await this._context.exposeBinding('_playwrightResume', false, () => {
            this._resume(false).catch(() => { });
        });
        await this._context.extendInjectedScript(recorderSource.source);
        await this._context.extendInjectedScript(consoleApiSource.source);
        this._context.recorderAppForTest = recorderApp;
    }
    async pause() {
        this._pauseDetails = { message: 'paused' };
        this._recorderApp.setPaused(this._pauseDetails);
        return new Promise(f => this._resumeCallback = f);
    }
    async _resume(step) {
        var _a;
        this._pauseOnNextStatement = step;
        if (this._resumeCallback)
            this._resumeCallback();
        this._resumeCallback = null;
        this._pauseDetails = null;
        (_a = this._recorderApp) === null || _a === void 0 ? void 0 : _a.setPaused(null);
    }
    async _onPage(page) {
        // First page is called page, others are called popup1, popup2, etc.
        const frame = page.mainFrame();
        page.on('close', () => {
            this._pageAliases.delete(page);
            this._generator.addAction({
                pageAlias,
                ...utils_1.describeFrame(page.mainFrame()),
                committed: true,
                action: {
                    name: 'closePage',
                    signals: [],
                }
            });
        });
        frame.on(frames_1.Frame.Events.Navigation, () => this._onFrameNavigated(frame, page));
        page.on(page_1.Page.Events.Download, () => this._onDownload(page));
        page.on(page_1.Page.Events.Popup, popup => this._onPopup(page, popup));
        page.on(page_1.Page.Events.Dialog, () => this._onDialog(page));
        const suffix = this._pageAliases.size ? String(++this._lastPopupOrdinal) : '';
        const pageAlias = 'page' + suffix;
        this._pageAliases.set(page, pageAlias);
        const isPopup = !!await page.opener();
        // Could happen due to the await above.
        if (page.isClosed())
            return;
        if (!isPopup) {
            this._generator.addAction({
                pageAlias,
                ...utils_1.describeFrame(page.mainFrame()),
                committed: true,
                action: {
                    name: 'openPage',
                    url: page.mainFrame().url(),
                    signals: [],
                }
            });
        }
    }
    _clearScript() {
        this._bufferedOutput.clear();
        this._generator.restart();
        if (!!this._params.startRecording) {
            for (const page of this._context.pages())
                this._onFrameNavigated(page.mainFrame(), page);
        }
    }
    async _performAction(frame, action) {
        const page = frame._page;
        const actionInContext = {
            pageAlias: this._pageAliases.get(page),
            ...utils_1.describeFrame(frame),
            action
        };
        this._generator.willPerformAction(actionInContext);
        const noCallMetadata = instrumentation_1.internalCallMetadata();
        try {
            const kActionTimeout = 5000;
            if (action.name === 'click') {
                const { options } = utils_1.toClickOptions(action);
                await frame.click(noCallMetadata, action.selector, { ...options, timeout: kActionTimeout });
            }
            if (action.name === 'press') {
                const modifiers = utils_1.toModifiers(action.modifiers);
                const shortcut = [...modifiers, action.key].join('+');
                await frame.press(noCallMetadata, action.selector, shortcut, { timeout: kActionTimeout });
            }
            if (action.name === 'check')
                await frame.check(noCallMetadata, action.selector, { timeout: kActionTimeout });
            if (action.name === 'uncheck')
                await frame.uncheck(noCallMetadata, action.selector, { timeout: kActionTimeout });
            if (action.name === 'select')
                await frame.selectOption(noCallMetadata, action.selector, [], action.options.map(value => ({ value })), { timeout: kActionTimeout });
        }
        catch (e) {
            this._generator.performedActionFailed(actionInContext);
            return;
        }
        const timer = setTimeout(() => {
            actionInContext.committed = true;
            this._timers.delete(timer);
        }, 5000);
        this._generator.didPerformAction(actionInContext);
        this._timers.add(timer);
    }
    async _recordAction(frame, action) {
        this._generator.addAction({
            pageAlias: this._pageAliases.get(frame._page),
            ...utils_1.describeFrame(frame),
            action
        });
    }
    _onFrameNavigated(frame, page) {
        const pageAlias = this._pageAliases.get(page);
        this._generator.signal(pageAlias, frame, { name: 'navigation', url: frame.url() });
    }
    _onPopup(page, popup) {
        const pageAlias = this._pageAliases.get(page);
        const popupAlias = this._pageAliases.get(popup);
        this._generator.signal(pageAlias, page.mainFrame(), { name: 'popup', popupAlias });
    }
    _onDownload(page) {
        const pageAlias = this._pageAliases.get(page);
        this._generator.signal(pageAlias, page.mainFrame(), { name: 'download' });
    }
    _onDialog(page) {
        const pageAlias = this._pageAliases.get(page);
        this._generator.signal(pageAlias, page.mainFrame(), { name: 'dialog', dialogAlias: String(++this._lastDialogOrdinal) });
    }
    async onBeforeCall(sdkObject, metadata) {
        var _a;
        this._sdkObject = sdkObject;
        this._callMetadata = metadata;
        const { source, line } = this._source(metadata);
        (_a = this._recorderApp) === null || _a === void 0 ? void 0 : _a.setSource(source, 'javascript', line);
        if (metadata.method === 'pause' || (this._pauseOnNextStatement && metadata.method === 'goto'))
            await this.pause();
    }
    async onAfterCall(sdkObject, metadata) {
        this._sdkObject = null;
        this._callMetadata = null;
    }
    async onBeforeInputAction(sdkObject, metadata) {
        if (this._pauseOnNextStatement)
            await this.pause();
    }
    _source(metadata) {
        let source = '// No source available';
        let line = undefined;
        if (metadata.stack && metadata.stack.length) {
            try {
                source = this._readAndCacheSource(metadata.stack[0].file);
                line = metadata.stack[0].line ? metadata.stack[0].line - 1 : undefined;
            }
            catch (e) {
                source = metadata.stack.join('\n');
            }
        }
        return { source, line };
    }
    _readAndCacheSource(fileName) {
        let source = this._sourceCache.get(fileName);
        if (source)
            return source;
        try {
            source = fs.readFileSync(fileName, 'utf-8');
        }
        catch (e) {
            source = '// No source available';
        }
        this._sourceCache.set(fileName, source);
        return source;
    }
}
exports.RecorderSupplement = RecorderSupplement;
//# sourceMappingURL=recorderSupplement.js.map